from __future__ import absolute_import

import io

import pandas as pd


def read_csv(csv_filepath, csv_meta=None):
    """
    Filters the rows of a CSV file using validator functions defined in a
    custom meta dictionary, and returns dicts (one per row).
    """
    with io.open(csv_filepath, 'r', encoding='utf-8') as f:
        df = pd.read_csv(io.StringIO(f.read()), float_precision='high')
        df = df.where(df.notnull(), None)

    for i in range(len(df)):
        r = df.iloc[i].to_dict()
        if not csv_meta:
            yield r
        else:
            yield {
                k: csv_meta[k]['validator'](r[csv_meta[k]['csv_header']]) for k in csv_meta
            }



# -*- coding: utf-8 -*-

__all__ = [
    'generate_files_for_reinsurance',
    'ReinsuranceLayer',
]

import pandas as pd
import os
import logging
import subprocess
import anytree
import shutil
import json
from collections import namedtuple

from oasislmf.exposures import oed


# Meta-data about an inuring layer
InuringLayer = namedtuple(
    "InuringLayer",
    "inuring_priority reins_numbers is_valid validation_messages")


def generate_files_for_reinsurance(
       # account_df,
       # location_df,
        items,
        coverages,
        fm_xrefs,
        xref_descriptions,
        ri_info_df,
        ri_scope_df,
        direct_oasis_files_dir):
    """
    Generate files for reinsurance.
    """

    inuring_metadata = {}
    previous_inuring_priority = None
    previous_risk_level = None
    reinsurance_index = 1
    for inuring_priority in range(1, ri_info_df['InuringPriority'].max() + 1):
        # Filter the reinsNumbers by inuring_priority
        reins_numbers = ri_info_df[ri_info_df['InuringPriority'] == inuring_priority].ReinsNumber.tolist()
        risk_level_set = set(ri_scope_df[ri_scope_df['ReinsNumber'].isin(reins_numbers)].RiskLevel)

        for risk_level in oed.REINS_RISK_LEVELS:
            if risk_level not in risk_level_set:
                continue

            written_to_dir = _generate_files_for_reinsurance_risk_level(
                inuring_priority,
                #account_df,
                #location_df,
                items,
                coverages,
                fm_xrefs,
                xref_descriptions,
                ri_info_df,
                ri_scope_df,
                previous_inuring_priority,
                previous_risk_level,
                risk_level,
                reinsurance_index,
                direct_oasis_files_dir)

            inuring_metadata[reinsurance_index] = {
                'inuring_priority': inuring_priority,
                'risk_level': risk_level,
                'directory': written_to_dir,
            }
            previous_inuring_priority = inuring_priority
            previous_risk_level = risk_level
            reinsurance_index = reinsurance_index + 1

    return inuring_metadata


def _generate_files_for_reinsurance_risk_level(
        inuring_priority,
        #account_df,
        #location_df,
        items,
        coverages,
        fm_xrefs,
        xref_descriptions,
        ri_info_df,
        ri_scope_df,
        previous_inuring_priority,
        previous_risk_level,
        risk_level,
        reinsurance_index,
        direct_oasis_files_dir):
    """
    Generate files for a reinsurance risk level.
    """
    reins_numbers_1 = ri_info_df[
        ri_info_df['InuringPriority'] == inuring_priority].ReinsNumber
    if reins_numbers_1.empty:
        return None
    reins_numbers_2 = ri_scope_df[
        ri_scope_df.isin({"ReinsNumber": reins_numbers_1.tolist()}).ReinsNumber &
        (ri_scope_df.RiskLevel == risk_level)].ReinsNumber
    if reins_numbers_2.empty:
        return None

    ri_info_inuring_priority_df = ri_info_df[ri_info_df.isin(
        {"ReinsNumber": reins_numbers_2.tolist()}).ReinsNumber]
    output_name = "ri_{}_{}".format(inuring_priority, risk_level)
    reinsurance_layer = ReinsuranceLayer(
        name=output_name,
        ri_info=ri_info_inuring_priority_df,
        ri_scope=ri_scope_df,
        #accounts=account_df,
        #locations=location_df,
        items=items,
        coverages=coverages,
        fm_xrefs=fm_xrefs,
        xref_descriptions=xref_descriptions,
        risk_level=risk_level
    )

    reinsurance_layer.generate_oasis_structures()
    #output_dir = os.path.join(direct_oasis_files_dir, output_name)
    output_dir = os.path.join(direct_oasis_files_dir, "RI_{}".format(reinsurance_index))
    reinsurance_layer.write_oasis_files(output_dir)
    return output_dir




class ReinsuranceLayer(object):
    """
    Generates ktools inputs and runs financial module for a reinsurance structure.
    """

    def __init__(self, name, ri_info, ri_scope, #accounts, locations,
                 items, coverages, fm_xrefs, xref_descriptions, risk_level, logger=None):

        self.logger = logger or logging.getLogger()
        self.name = name
        #self.accounts = accounts
        #self.locations = locations

        self.coverages = coverages
        self.items = items
        self.fm_xrefs = fm_xrefs
        self.xref_descriptions = xref_descriptions

        self.item_ids = list()
        self.item_tivs = list()
        self.fmprogrammes = pd.DataFrame()
        self.fmprofiles = pd.DataFrame()
        self.fm_policytcs = pd.DataFrame()

        self.risk_level = risk_level

        self.ri_info = ri_info
        self.ri_scope = ri_scope

        self.add_profiles_args = namedtuple(
            "AddProfilesArgs",
            "program_node, ri_info_row, scope_rows, overlay_loop, layer_id, "
            "node_layer_profile_map, fmprofiles_list, nolossprofile_id, passthroughprofile_id")

    def _add_program_node(self, level_id):
        return anytree.Node(
            "Treaty",
            parent=None,
            level_id=level_id,
            agg_id=1,
            account_number=oed.NOT_SET_ID,
            policy_number=oed.NOT_SET_ID,
            location_number=oed.NOT_SET_ID)

    def _add_item_node(self, xref_id, parent):
        return anytree.Node(
            "Item_id:{}".format(xref_id),
            parent=parent,
            level_id=1,
            agg_id=xref_id,
            account_number=oed.NOT_SET_ID,
            policy_number=oed.NOT_SET_ID,
            location_number=oed.NOT_SET_ID)

    def _add_location_node(
            self, level_id, agg_id, xref_description, parent):
        return anytree.Node(
            "Account_number:{} Policy_number:{} Location_number:{}".format(
                xref_description.account_number,
                xref_description.policy_number,
                xref_description.location_number),
            parent=parent,
            level_id=level_id,
            agg_id=agg_id,
            account_number=xref_description.account_number,
            policy_number=xref_description.policy_number,
            location_number=xref_description.location_number)

    def _add_policy_node(
            self, level_id, agg_id, xref_description, parent):
        return anytree.Node(
            "Account_number:{} Policy_number:{}".format(
                xref_description.account_number, xref_description.policy_number),
            parent=parent,
            level_id=level_id,
            agg_id=agg_id,
            account_number=xref_description.account_number,
            policy_number=xref_description.policy_number,
            location_number=oed.NOT_SET_ID)

    def _add_account_node(
            self, agg_id, level_id, xref_description, parent):
        return anytree.Node(
            "Account_number:{}".format(xref_description.account_number),
            parent=parent,
            level_id=level_id,
            agg_id=agg_id,
            account_number=xref_description.account_number,
            policy_number=oed.NOT_SET_ID,
            location_number=oed.NOT_SET_ID)



    def _does_location_node_match_scope_row(self, node, ri_scope_row):
        node_summary = (node.account_number,
                        node.policy_number, node.location_number)
        scope_row_summary = (ri_scope_row.AccNumber,
                             ri_scope_row.PolNumber, ri_scope_row.LocNumber)
        if (node_summary == scope_row_summary):
            self.logger.debug('Matching node: location to scope\n node: {}, ri_scope: {}'.format(
                str(node_summary),
                str(scope_row_summary),
            ))
        return (node_summary == scope_row_summary)

    def _does_policy_node_match_scope_row(self, node, ri_scope_row):
        node_summary = (node.account_number,
                        node.policy_number, oed.NOT_SET_ID)
        scope_row_summary = (ri_scope_row.AccNumber, ri_scope_row.PolNumber, oed.NOT_SET_ID)
        if (node_summary == scope_row_summary):
            self.logger.debug('Matching node: policy to scope\n node: {}, ri_scope: {}'.format(
                str(node_summary),
                str(scope_row_summary),
            ))
        return (node_summary == scope_row_summary)

    def _does_account_node_match_scope_row(self, node, ri_scope_row):
        node_summary = (node.account_number,
                        oed.NOT_SET_ID, oed.NOT_SET_ID)
        scope_row_summary = (ri_scope_row.AccNumber,
                             oed.NOT_SET_ID, oed.NOT_SET_ID)
        if (node_summary == scope_row_summary):
            self.logger.debug('Matching node: account to scope\n node: {}, ri_scope: {}'.format(
                str(node_summary),
                str(scope_row_summary),
            ))
        return (node_summary == scope_row_summary)

    ## More generic but slower (testing only)
    #def _match_node(self, node, search_dict):
    #    node_dict = {
    #        'AccNumber':  node.account_number,
    #        'PolNumber':   node.policy_number,
    #        'LocNumber': node.location_number,
    #    }
    #    self.logger.debug('Matching node: \n\t node: {}, \n\t search: {}'.format(
    #        str(node_dict),
    #        str(search_dict),
    #    ))
    #    return search_dict.items() <= node_dict.items()

    def _filter_nodes(self, nodes_list, scope_row):
        """
        Return subset of `nodes_list` based on values of a row in `ri_scope.csv`

        TODO: Combined filters?
        """
        if (scope_row.RiskLevel == oed.REINS_RISK_LEVEL_ACCOUNT) and self._is_defined(scope_row.AccNumber):
            return list(filter(lambda n: n.account_number == scope_row.AccNumber, nodes_list))
        elif scope_row.RiskLevel == oed.REINS_RISK_LEVEL_LOCATION and self._is_defined(scope_row.LocNumber):
            return list(filter(lambda n: n.location_number == scope_row.LocNumber, nodes_list))
        elif scope_row.RiskLevel == oed.REINS_RISK_LEVEL_POLICY and self._is_defined(scope_row.PolNumber):
            return list(filter(lambda n: n.policy_number == scope_row.PolNumber, nodes_list))
        elif scope_row.RiskLevel == oed.REINS_RISK_LEVEL_PORTFOLIO and self._is_defined(scope_row.PortNumber):
            return list(filter(lambda n: n.policy_number == scope_row.PortNumber, nodes_list))
        else:
            return nodes_list



    def _is_defined(self, num_to_check):
        # If the value = NaN it will return False
        return num_to_check == num_to_check


    def _get_tree(self):
        current_location_number = 0
        current_policy_number = 0
        current_account_number = 0
        current_location_node = None
        current_policy_node = None
        current_account_node = None

        program_node_level_id = 3
        if self.risk_level == oed.REINS_RISK_LEVEL_PORTFOLIO:
            program_node_level_id = 2

        program_node = self._add_program_node(program_node_level_id)

        xref_descriptions = self.xref_descriptions.sort_values(
            by=["location_number", "policy_number", "account_number"])
        agg_id = 0
        if self.risk_level == oed.REINS_RISK_LEVEL_PORTFOLIO:
            for _, row in xref_descriptions.iterrows():
                self._add_item_node(row.xref_id, program_node)
        elif self.risk_level == oed.REINS_RISK_LEVEL_ACCOUNT:
            for _, row in xref_descriptions.iterrows():
                if current_account_number != row.account_number:
                    agg_id = agg_id + 1
                    level_id = 2
                    current_account_number = row.account_number
                    current_account_node = self._add_account_node(
                        agg_id, level_id, row, program_node)
                self._add_item_node(row.xref_id, current_account_node)
        elif self.risk_level == oed.REINS_RISK_LEVEL_POLICY:
            for _, row in xref_descriptions.iterrows():
                if current_policy_number != row.policy_number:
                    agg_id = agg_id + 1
                    level_id = 2
                    current_policy_number = row.policy_number
                    current_policy_node = self._add_location_node(
                        level_id, agg_id, row, program_node)
                self._add_item_node(row.xref_id, current_policy_node)
        elif self.risk_level == oed.REINS_RISK_LEVEL_LOCATION:
            for _, row in xref_descriptions.iterrows():
                if current_location_number != row.location_number:
                    agg_id = agg_id + 1
                    level_id = 2
                    current_location_number = row.location_number
                    current_location_node = self._add_location_node(
                        level_id, agg_id, row, program_node)
                self._add_item_node(row.xref_id, current_location_node)
        return program_node

#    def _add_occ_limit(self, add_profiles_args):



    def _add_fac_profiles(self, add_profiles_args):
        self.logger.debug("Adding FAC profiles:")
        profile_id = max(
            x.profile_id for x in add_profiles_args.fmprofiles_list)

        profile_id = profile_id + 1
        add_profiles_args.fmprofiles_list.append(oed.get_reinsurance_profile(
            profile_id,
            attachment=add_profiles_args.ri_info_row.RiskAttachment,
            limit=add_profiles_args.ri_info_row.RiskLimit,
            ceded=add_profiles_args.ri_info_row.CededPercent,
            placement=add_profiles_args.ri_info_row.PlacedPercent
        ))

        for _, ri_scope_row in add_profiles_args.scope_rows.iterrows():
            if ri_scope_row.RiskLevel == oed.REINS_RISK_LEVEL_LOCATION:
                nodes = anytree.search.findall(
                    add_profiles_args.program_node,
                    filter_=lambda node: self._does_location_node_match_scope_row(node, ri_scope_row))
                for node in nodes:
                    add_profiles_args.node_layer_profile_map[(
                        node.name, add_profiles_args.layer_id, add_profiles_args.overlay_loop)] = profile_id
            elif ri_scope_row.RiskLevel == oed.REINS_RISK_LEVEL_POLICY:
                nodes = anytree.search.findall(
                    add_profiles_args.program_node,
                    filter_=lambda node: self._does_policy_node_match_scope_row(node, ri_scope_row))
                for node in nodes:
                    add_profiles_args.node_layer_profile_map[(
                        node.name, add_profiles_args.layer_id, add_profiles_args.overlay_loop)] = profile_id
            elif ri_scope_row.RiskLevel == oed.REINS_RISK_LEVEL_ACCOUNT:
                nodes = anytree.search.findall(
                    add_profiles_args.program_node,
                    filter_=lambda node: self._does_account_node_match_scope_row(node, ri_scope_row))
                for node in nodes:
                    add_profiles_args.node_layer_profile_map[(
                        node.name, add_profiles_args.layer_id, add_profiles_args.overlay_loop)] = profile_id
            else:
                raise Exception(
                    "Unsupported risk level: {}".format(ri_scope_row.RiskLevel))

    # Need to check Matching rules for Per Risk with Joh
    def _add_per_risk_profiles(self, add_profiles_args):
        self.logger.debug("Adding PR profiles:")
        profile_id = max(x.profile_id for x in add_profiles_args.fmprofiles_list)
        nodes_all = anytree.search.findall(
               add_profiles_args.program_node, filter_=lambda node: node.level_id == 2)

        profile_id = profile_id + 1
        add_profiles_args.fmprofiles_list.append(oed.get_reinsurance_profile(
            profile_id,
            attachment=add_profiles_args.ri_info_row.RiskAttachment,
            limit=add_profiles_args.ri_info_row.RiskLimit,
            ceded=add_profiles_args.ri_info_row.CededPercent,
        ))

        for _, ri_scope_row in add_profiles_args.scope_rows.iterrows():
            selected_nodes = self._filter_nodes(nodes_all, ri_scope_row)
            for node in selected_nodes:
                add_profiles_args.node_layer_profile_map[(
                    node.name, add_profiles_args.layer_id, add_profiles_args.overlay_loop)] = profile_id

        # add OccLimit / Placed Percent
        profile_id = profile_id + 1
        add_profiles_args.fmprofiles_list.append(
            oed.get_occlim_profile(
                profile_id,
                limit=add_profiles_args.ri_info_row.OccLimit,
                placement=add_profiles_args.ri_info_row.PlacedPercent,
        ))
        add_profiles_args.node_layer_profile_map[
            (add_profiles_args.program_node.name, add_profiles_args.layer_id, add_profiles_args.overlay_loop)] = profile_id





    def _add_surplus_share_profiles(self, add_profiles_args):
        self.logger.debug("Adding SS profiles:")
        profile_id = max( x.profile_id for x in add_profiles_args.fmprofiles_list)
        nodes_all = anytree.search.findall(
               add_profiles_args.program_node, filter_=lambda node: node.level_id == 2)

        for _, ri_scope_row in add_profiles_args.scope_rows.iterrows():
            profile_id = profile_id + 1
            add_profiles_args.fmprofiles_list.append(oed.get_reinsurance_profile(
                profile_id,
                attachment=add_profiles_args.ri_info_row.RiskAttachment,
                limit=add_profiles_args.ri_info_row.RiskLimit,
                ceded=ri_scope_row.CededPercent,
            ))

            if ri_scope_row.RiskLevel == oed.REINS_RISK_LEVEL_LOCATION:
                selected_nodes = list(filter(lambda n: self._does_location_node_match_scope_row(n,ri_scope_row), nodes_all))
                for node in selected_nodes:
                    add_profiles_args.node_layer_profile_map[(
                        node.name, add_profiles_args.layer_id, add_profiles_args.overlay_loop)] = profile_id
            elif ri_scope_row.RiskLevel == oed.REINS_RISK_LEVEL_POLICY:
                selected_nodes = list(filter(lambda n: self._does_policy_node_match_scope_row(n,ri_scope_row), nodes_all))
                for node in selected_nodes:
                    add_profiles_args.node_layer_profile_map[(
                        node.name, add_profiles_args.layer_id, add_profiles_args.overlay_loop)] = profile_id
            elif ri_scope_row.RiskLevel == oed.REINS_RISK_LEVEL_ACCOUNT:
                selected_nodes = list(filter(lambda n: self._does_account_node_match_scope_row(n,ri_scope_row), nodes_all))
                for node in selected_nodes:
                    add_profiles_args.node_layer_profile_map[(
                        node.name, add_profiles_args.layer_id, add_profiles_args.overlay_loop)] = profile_id
            else:
                raise Exception(
                    "Unsupported risk level: {}".format(ri_scope_row.RiskLevel))


        # add OccLimit / Placed Percent
        profile_id = profile_id + 1
        add_profiles_args.fmprofiles_list.append(
            oed.get_occlim_profile(
                profile_id,
                limit=add_profiles_args.ri_info_row.OccLimit,
                placement=add_profiles_args.ri_info_row.PlacedPercent,
        ))
        add_profiles_args.node_layer_profile_map[
            (add_profiles_args.program_node.name, add_profiles_args.layer_id, add_profiles_args.overlay_loop)] = profile_id




    def _add_quota_share_profiles(self, add_profiles_args):
        self.logger.debug("Adding QS profiles:")

        profile_id = max( x.profile_id for x in add_profiles_args.fmprofiles_list)
        nodes_all = anytree.search.findall(
               add_profiles_args.program_node, filter_=lambda node: node.level_id == 2)

        # Add any risk limits
        # RISK LEVEL SEL
        if self.risk_level == oed.REINS_RISK_LEVEL_PORTFOLIO:
            profile_id = profile_id + 1
            add_profiles_args.fmprofiles_list.append(
                oed.get_reinsurance_profile(
                    profile_id,
                    limit=add_profiles_args.ri_info_row.OccLimit,
                    ceded=add_profiles_args.ri_info_row.CededPercent,
                    placement=add_profiles_args.ri_info_row.PlacedPercent
            ))
        else:
            for _, ri_scope_row in add_profiles_args.scope_rows.iterrows():
                profile_id = profile_id + 1

                add_profiles_args.fmprofiles_list.append(
                    oed.get_reinsurance_profile(
                        profile_id,
                        limit=add_profiles_args.ri_info_row.RiskLimit,
                        ceded=add_profiles_args.ri_info_row.CededPercent,
                    ))

                # Filter
                selected_nodes = self._filter_nodes(nodes_all, ri_scope_row)
                for node in selected_nodes:
                    add_profiles_args.node_layer_profile_map[(
                        node.name, add_profiles_args.layer_id, add_profiles_args.overlay_loop)] = profile_id

        # add OccLimit / Placed Percent
        profile_id = profile_id + 1
        add_profiles_args.fmprofiles_list.append(
            oed.get_occlim_profile(
                profile_id,
                limit=add_profiles_args.ri_info_row.OccLimit,
                placement=add_profiles_args.ri_info_row.PlacedPercent,
        ))
        add_profiles_args.node_layer_profile_map[
            (add_profiles_args.program_node.name, add_profiles_args.layer_id, add_profiles_args.overlay_loop)] = profile_id

    def _add_cat_xl_profiles(self, add_profiles_args):
        self.logger.debug("Adding CAT XL profiles:")
        profile_id = max(
            x.profile_id for x in add_profiles_args.fmprofiles_list)

        if self.risk_level == oed.REINS_RISK_LEVEL_PORTFOLIO:
            profile_id = profile_id + 1
            add_profiles_args.fmprofiles_list.append(
                oed.get_reinsurance_profile(
                    profile_id,
                    attachment=add_profiles_args.ri_info_row.OccAttachment,
                    limit=add_profiles_args.ri_info_row.OccLimit,
                    placement=add_profiles_args.ri_info_row.PlacedPercent,
                    ceded=add_profiles_args.ri_info_row.CededPercent,
                ))
            add_profiles_args.node_layer_profile_map[
                (add_profiles_args.program_node.name, add_profiles_args.layer_id, add_profiles_args.overlay_loop)] = profile_id

        else:
            nodes_all = anytree.search.findall(
                   add_profiles_args.program_node, filter_=lambda node: node.level_id == 2)

            for _, ri_scope_row in add_profiles_args.scope_rows.iterrows():
                profile_id = profile_id + 1

                add_profiles_args.fmprofiles_list.append(
                    oed.get_pass_through_profile(
                        profile_id,
                    )
                )

                # Filter
                selected_nodes = self._filter_nodes(nodes_all, ri_scope_row)
                for node in selected_nodes:
                    add_profiles_args.node_layer_profile_map[(
                        node.name, add_profiles_args.layer_id, add_profiles_args.overlay_loop)] = profile_id

            # add OccLimit / Placed Percent
            profile_id = profile_id + 1
            add_profiles_args.fmprofiles_list.append(
                oed.get_reinsurance_profile(
                    profile_id,
                    attachment=add_profiles_args.ri_info_row.OccAttachment,
                    ceded=add_profiles_args.ri_info_row.CededPercent,
                    limit=add_profiles_args.ri_info_row.OccLimit,
                    placement=add_profiles_args.ri_info_row.PlacedPercent,
            ))
            add_profiles_args.node_layer_profile_map[
                (add_profiles_args.program_node.name, add_profiles_args.layer_id, add_profiles_args.overlay_loop)] = profile_id

    def generate_oasis_structures(self):
        '''
        Create the Oasis structures - FM Programmes, FM Profiles and FM Policy TCs -
        that represent the resinsurance structure.

        The algorithm to create the stucture has three steps:
        Step 1 - Build a tree representation of the insurance program, depening on the reinsuarnce risk level.
        Step 2 - Overlay the reinsurance structure. Each resinsuarnce contact is a seperate layer.
        Step 3 - Iterate over the tree and write out the Oasis structure.
        '''

        fmprogrammes_list = list()
        fmprofiles_list = list()
        fm_policytcs_list = list()

        profile_id = 1
        nolossprofile_id = profile_id
        fmprofiles_list.append(
            oed.get_no_loss_profile(nolossprofile_id))
        profile_id = profile_id + 1
        passthroughprofile_id = profile_id
        fmprofiles_list.append(
            oed.get_pass_through_profile(passthroughprofile_id))

        node_layer_profile_map = {}

        self.logger.debug(fmprofiles_list)


        #
        # Step 1 - Build a tree representation of the insurance program, depening on the reinsuarnce risk level.
        #
        program_node = self._get_tree()



        if self.logger:
            self.logger.debug('program_node tree: "{}"'.format(self.name))
            self.logger.debug(anytree.RenderTree(program_node))
            #Plot tree to image (graphviz)
            #from anytree.dotexport import RenderTreeGraph
            #RenderTreeGraph(program_node).to_picture(
            #    "Init_{}.png".format(self.name))


        #
        # Step 2 - Overlay the reinsurance structure. Each resinsuarnce contact is a seperate layer.
        #
        layer_id = 1        # Current layer ID
        overlay_loop = 0    # Overlays multiple rules in same layer
        prev_reins_number = 1
        for _, ri_info_row in self.ri_info.iterrows():
            overlay_loop += 1
            scope_rows = self.ri_scope[
                (self.ri_scope.ReinsNumber == ri_info_row.ReinsNumber) &
                (self.ri_scope.RiskLevel == self.risk_level)]

            # Three rules for layers
            # 1. if FAC don't inncrement the layer number
            # 2. Otherwise, only increment inline with the reins_number
            # 3. If the reins_number number is the same as prev, increment based on ReinsLayerNumber
            if ri_info_row.ReinsType in ['FAC']:
                pass
            elif prev_reins_number < ri_info_row.ReinsNumber:
                layer_id += 1
                prev_reins_number = ri_info_row.ReinsNumber
            elif layer_id < ri_info_row.ReinsLayerNumber:
                layer_id = ri_info_row.ReinsLayerNumber


            if self.logger:
                pd.set_option('display.width', 1000)
                self.logger.debug('ri_scope: "{}"'.format(self.name))
                self.logger.debug(scope_rows)

            if scope_rows.shape[0] == 0:
                continue

            add_profiles_args = self.add_profiles_args(
                program_node, ri_info_row, scope_rows, overlay_loop, layer_id,
                node_layer_profile_map, fmprofiles_list,
                nolossprofile_id, passthroughprofile_id)

            # Add pass through nodes at all levels so that the risks
            # not explicitly covered are unaffected
            for node in anytree.iterators.LevelOrderIter(add_profiles_args.program_node):
                add_profiles_args.node_layer_profile_map[(
                    node.name, add_profiles_args.layer_id, add_profiles_args.overlay_loop)] = add_profiles_args.nolossprofile_id
            add_profiles_args.node_layer_profile_map[(
                add_profiles_args.program_node.name, add_profiles_args.layer_id, add_profiles_args.overlay_loop)] = add_profiles_args.passthroughprofile_id

            if ri_info_row.ReinsType == oed.REINS_TYPE_FAC:
                self._add_fac_profiles(add_profiles_args)
            elif ri_info_row.ReinsType == oed.REINS_TYPE_PER_RISK:
                self._add_per_risk_profiles(add_profiles_args)
            elif ri_info_row.ReinsType == oed.REINS_TYPE_QUOTA_SHARE:
                self._add_quota_share_profiles(add_profiles_args)
            elif ri_info_row.ReinsType == oed.REINS_TYPE_SURPLUS_SHARE:
                self._add_surplus_share_profiles(add_profiles_args)
            elif ri_info_row.ReinsType == oed.REINS_TYPE_CAT_XL:
                self._add_cat_xl_profiles(add_profiles_args)
            else:
                raise Exception("ReinsType not supported yet: {}".format(
                    ri_info_row.ReinsType))


        #
        # Step 3 - Iterate over the tree and write out the Oasis structure.
        #
        for node in anytree.iterators.LevelOrderIter(program_node):
            if node.parent is not None:
                fmprogrammes_list.append(
                    oed.FmProgramme(
                        from_agg_id=node.agg_id,
                        level_id=node.level_id,
                        to_agg_id=node.parent.agg_id
                    )
                )

        # Note: Pending confirmation from Joh that ReinsLayerNumber is being used correctly
        for layer in range(1,layer_id+1):
            for node in anytree.iterators.LevelOrderIter(program_node):
                if node.level_id > 1:
                    profiles_ids = []

                    # The `overlay_rule` replaces using each resinsuarnce contact in a seperate layer
                    # Collect overlaping unique combinations of (layer_id, level_id, agg_id) and combine into
                    # a single layer
                    for overlay_rule in range(1,overlay_loop+1):
                        try:
                            profiles_ids.append(
                                node_layer_profile_map[(node.name, layer, overlay_rule)])

                        except:
                            profiles_ids.append(1)
                            pass
                    fm_policytcs_list.append(oed.FmPolicyTc(
                        layer_id=layer,
                        level_id=node.level_id - 1,
                        agg_id=node.agg_id,
                        profile_id=max(profiles_ids)
                    ))
        self.fmprogrammes = pd.DataFrame(fmprogrammes_list)
        self.fmprofiles = pd.DataFrame(fmprofiles_list)
        self.fm_policytcs = pd.DataFrame(fm_policytcs_list)
        self.fm_xrefs['layer_id'] = pd.Series(layer_id, range(len(self.fm_xrefs.index)))


        # Log Reinsurance structures
        if self.logger:
            self.logger.debug('policytc_map: "{}"'.format(self.name))
            policytc_map = dict()
            for k in add_profiles_args.node_layer_profile_map.keys():
                profile_id = add_profiles_args.node_layer_profile_map[k]
                policytc_map["(Name=%s, layer_id=%s, overlay_loop=%s)" % k] = profile_id
            self.logger.debug(json.dumps(policytc_map, indent=4))
            self.logger.debug('fm_policytcs: "{}"'.format(self.name))
            self.logger.debug(self.fm_policytcs)
            self.logger.debug('fm_profile: "{}"'.format(self.name))
            self.logger.debug(self.fmprofiles)

    def write_oasis_files(self, directory=None):

        if directory is None:
            directory = "direct"
        if os.path.exists(directory):
            shutil.rmtree(directory)
        os.makedirs(directory)

        self.coverages.to_csv(
            os.path.join(directory, "coverages.csv"), index=False)
        self.items.to_csv(
            os.path.join(directory, "items.csv"), index=False)
        self.fmprogrammes.to_csv(
            os.path.join(directory, "fm_programme.csv"), index=False)
        self.fmprofiles.to_csv(
            os.path.join(directory, "fm_profile.csv"), index=False)
        self.fm_policytcs.to_csv(
            os.path.join(directory, "fm_policytc.csv"), index=False)
        self.fm_xrefs.to_csv(
            os.path.join(directory, "fm_xref.csv"), index=False)
        open(os.path.join(directory, "fmsummaryxref.csv"), 'a').close()
        open(os.path.join(directory, "gulsummaryxref.csv"), 'a').close()



import json
import logging

import sys

import os
import io

from argparsetree import BaseCommand

from oasislmf.utils.exceptions import OasisException
from .cleaners import PathCleaner


class InputValues(object):
    """
    Helper class for accessing the input values from either
    the command line or the configuration file.
    """
    def __init__(self, args):
        self.args = args

        self.config = {}
        self.config_dir = os.path.dirname(args.config)
        if os.path.exists(args.config):
            with io.open(args.config, 'r', encoding='utf-8') as f:
                self.config = json.load(f)

    def get(self, name, default=None, required=False, is_path=False):
        """
        Gets the names parameter from the command line arguments.

        If it is not set on the command line the configuration file
        is checked.

        If it is also not present in the configuration file then
        ``default`` is returned unless ``required`` is false in which
        case an ``OasisException`` is raised.

        :param name: The name of the parameter to lookup
        :type name: str

        :param default: The default value to return if the name is not
            found on the command line or in the configuration file.

        :param required: Flag whether the value is required, if so and
            the parameter is not found on the command line or in the
            configuration file an error is raised.
        :type required: bool

        :param is_path: Flag whether the value should be treated as a path,
            is so the value is processed as relative to the config file.
        :type is_path: bool

        :raise OasisException: If the value is not found and ``required``
            is True

        :return: The found value or the default
        """
        value = None
        cmd_value = getattr(self.args, name, None)
        if cmd_value is not None:
            value = cmd_value
        elif name in self.config:
            value = self.config[name]

        if required and value is None:
            raise OasisException(
                '{} could not be found in the command args or config file ({}) but is required'.format(name, self.args.config)
            )

        if value is None:
            value = default

        if is_path and value is not None:
            p = os.path.join(self.config_dir, value)
            value = os.path.abspath(p) if not os.path.isabs(value) else p

        return value


class OasisBaseCommand(BaseCommand):
    """
    The base command to inherit from for each command.

    2 additional arguments (``--verbose`` and ``--config``) are added to
    the parser so that they are available for all commands.
    """
    def __init__(self, *args, **kwargs):
        self._logger = None
        super(OasisBaseCommand, self).__init__(*args, **kwargs)

    def add_args(self, parser):
        """
        Adds arguments to the argument parser. This is used to modify
        which arguments are processed by the command.

        2 global parameters (``--verbose`` and ``--config``) are added
        so that they are available to all commands.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """
        parser.add_argument('-V', '--verbose', action='store_true', help='Use verbose logging.')
        parser.add_argument(
            '-C', '--config', type=PathCleaner('Config file', preexists=False),
            help='The oasislmf config to load', default='./oasislmf.json'
        )

    def parse_args(self):
        """
        Parses the command line arguments and sets them in ``self.args``

        :return: The arguments taken from the command line
        """
        try:
            self.args = super(OasisBaseCommand, self).parse_args()
            self.setup_logger(self.args.verbose)
            return self.args
        except Exception:
            self.setup_logger(False)
            raise

    def setup_logger(self, verbose):
        """
        The logger to use for the command with the verbosity set
        """
        if not self._logger:
            if verbose:
                log_level = logging.DEBUG
                log_format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
            else:
                log_level = logging.INFO
                log_format = '%(message)s'

            logging.basicConfig(stream=sys.stdout, level=log_level, format=log_format)
            self._logger = logging.getLogger()

    @property
    def logger(self):
        if self._logger:
            return self._logger

        return logging.getLogger()



# -*- coding: utf-8 -*-

__all__ = [
    'get_dataframe'
]

import builtins

import pandas as pd

import six

from .exceptions import OasisException

def get_dataframe(
    src_fp=None,
    src_type='csv',
    src_buf=None,
    src_data=None,
    float_precision='high',
    lowercase_cols=True,
    index_col=True,
    non_na_cols=(),
    col_dtypes={},
    sort_col=None,
    sort_ascending=None
):
    if not (src_fp or src_buf or src_data is not None):
        raise OasisException(
            'A CSV or JSON file path or a string buffer of such a file or an '
            'appropriate data structure or dataframe must be provided'
        )

    df = None

    if src_fp and src_type == 'csv':
        df = pd.read_csv(src_fp, float_precision=float_precision)
    elif src_buf and src_type == 'csv':
        df = pd.read_csv(io.StringIO(src_buf), float_precision=float_precision)
    elif src_fp and src_type == 'json':
        df = pd.read_json(src_fp, precise_float=(True if float_precision == 'high' else False))
    elif src_buf and src_type == 'json':
        df = pd.read_json(io.StringIO(src_buf), precise_float=(True if float_precision == 'high' else False))
    elif src_data and (isinstance(src_data, list) or isinstance(src_data, pd.DataFrame)):
        df = pd.DataFrame(data=src_data, dtype=object)

    if lowercase_cols:
        df.columns = df.columns.str.lower()

    if index_col:
        df['index'] = list(range(len(df)))

    if non_na_cols:
        _non_na_cols = tuple(col.lower() for col in non_na_cols) if lowercase_cols else non_na_cols
        df.dropna(subset=_non_na_cols, inplace=True)

    if col_dtypes:
        _col_dtypes = {
            (k.lower() if lowercase_cols else k):(getattr(builtins, v) if v in ('int', 'bool', 'float', 'str',) else v) for k, v in six.iteritems(col_dtypes)
        }
        for col, dtype in six.iteritems(_col_dtypes):
            df[col] = df[col].astype(dtype) if dtype != int else df[col].astype(object)

    if sort_col:
        _sort_col = sort_col.lower() if lowercase_cols else sort_col
        sort_ascending = sort_ascending if sort_ascending is not None else True
        df.sort_values(_sort_col, axis=0, ascending=sort_ascending, inplace=True)

    return df



from __future__ import unicode_literals

import hashlib
import json
import os
import io
import shutil
from unittest import TestCase

from oasislmf.model_execution.bash import genbash
from oasislmf.utils import diff

TEST_DIRECTORY = os.path.dirname(__file__)

KPARSE_INPUT_FOLDER = os.path.join(TEST_DIRECTORY, "kparse_input")
KPARSE_OUTPUT_FOLDER = os.path.join(TEST_DIRECTORY, "kparse_output")
KPARSE_REFERENCE_FOLDER = os.path.join(TEST_DIRECTORY, "kparse_reference")


class Genbash(TestCase):
    @classmethod
    def setUpClass(cls):
        if os.path.exists(KPARSE_OUTPUT_FOLDER):
            shutil.rmtree(KPARSE_OUTPUT_FOLDER)
        os.makedirs(KPARSE_OUTPUT_FOLDER)

    def md5(self, fname):
        hash_md5 = hashlib.md5()
        with io.open(fname, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()

    def genbash(self, name, num_partitions, num_reinsurance_iterations=0):
        input_filename = os.path.join(KPARSE_INPUT_FOLDER, "{}.json".format(name))
        if num_reinsurance_iterations <= 0:
            output_filename = os.path.join(KPARSE_OUTPUT_FOLDER, "{}_{}_partition.sh".format(name, num_partitions))
        else:
            output_filename = os.path.join(
                KPARSE_OUTPUT_FOLDER, 
                "{}_{}_reins_layer_{}_partition.sh".format(name, num_reinsurance_iterations, num_partitions))
            
        with io.open(input_filename, encoding='utf-8') as file:
            analysis_settings = json.load(file)['analysis_settings']

        genbash(
            num_partitions,
            analysis_settings,
            output_filename,
            num_reinsurance_iterations
        )

    def check(self, name):
        output_filename = os.path.join(KPARSE_OUTPUT_FOLDER, "{}.sh".format(name))
        reference_filename = os.path.join(KPARSE_REFERENCE_FOLDER, "{}.sh".format(name))

        d = diff.unified_diff(reference_filename, output_filename, as_string=True)
        if d:
            self.fail(d)

    def test_gul_summarycalc_1_partition(self):
        self.genbash("gul_summarycalc_1_output", 1)
        self.check("gul_summarycalc_1_output_1_partition")

    def test_gul_summarycalc_20_partition(self):
        self.genbash("gul_summarycalc_1_output", 20)
        self.check("gul_summarycalc_1_output_20_partition")

    def test_gul_eltcalc_1_partition(self):
        self.genbash("gul_eltcalc_1_output", 1)
        self.check("gul_eltcalc_1_output_1_partition")

    def test_gul_eltcalc_20_partition(self):
        self.genbash("gul_eltcalc_1_output", 20)
        self.check("gul_eltcalc_1_output_20_partition")

    def test_gul_aalcalc_1_partition(self):
        self.genbash("gul_aalcalc_1_output", 1)
        self.check("gul_aalcalc_1_output_1_partition")

    def test_gul_aalcalc_20_partition(self):
        self.genbash("gul_aalcalc_1_output", 20)
        self.check("gul_aalcalc_1_output_20_partition")

    def test_gul_pltcalc_1_partition(self):
        self.genbash("gul_pltcalc_1_output", 1)
        self.check("gul_pltcalc_1_output_1_partition")

    def test_gul_pltcalc_20_partition(self):
        self.genbash("gul_pltcalc_1_output", 20)
        self.check("gul_pltcalc_1_output_20_partition")

    def test_gul_agg_fu_lec_1_partition(self):
        self.genbash("gul_agg_fu_lec_1_output", 1)
        self.check("gul_agg_fu_lec_1_output_1_partition")

    def test_gul_agg_fu_lec_20_partition(self):
        self.genbash("gul_agg_fu_lec_1_output", 20)
        self.check("gul_agg_fu_lec_1_output_20_partition")

    def test_gul_occ_fu_lec_1_output_1_partition(self):
        self.genbash("gul_occ_fu_lec_1_output", 1)
        self.check("gul_occ_fu_lec_1_output_1_partition")

    def test_gul_occ_fu_lec_1_output_20_partition(self):
        self.genbash("gul_occ_fu_lec_1_output", 20)
        self.check("gul_occ_fu_lec_1_output_20_partition")

    def test_gul_agg_ws_lec_1_partition(self):
        self.genbash("gul_agg_ws_lec_1_output", 1)
        self.check("gul_agg_ws_lec_1_output_1_partition")

    def test_gul_agg_ws_lec_20_partition(self):
        self.genbash("gul_agg_ws_lec_1_output", 20)
        self.check("gul_agg_ws_lec_1_output_20_partition")

    def test_gul_occ_ws_lec_1_partition(self):
        self.genbash("gul_occ_ws_lec_1_output", 1)
        self.check("gul_occ_ws_lec_1_output_1_partition")

    def test_gul_occ_ws_lec_20_partition(self):
        self.genbash("gul_occ_ws_lec_1_output", 20)
        self.check("gul_occ_ws_lec_1_output_20_partition")

    def test_gul_agg_ws_mean_lec_1_partition(self):
        self.genbash("gul_agg_ws_mean_lec_1_output", 1)
        self.check("gul_agg_ws_mean_lec_1_output_1_partition")

    def test_gul_agg_ws_mean_lec_20_partition(self):
        self.genbash("gul_agg_ws_mean_lec_1_output", 20)
        self.check("gul_agg_ws_mean_lec_1_output_20_partition")

    def test_gul_occ_ws_mean_lec_1_partition(self):
        self.genbash("gul_occ_ws_mean_lec_1_output", 1)
        self.check("gul_occ_ws_mean_lec_1_output_1_partition")

    def test_gul_occ_ws_mean_lec_20_partition(self):
        self.genbash("gul_occ_ws_mean_lec_1_output", 20)
        self.check("gul_occ_ws_mean_lec_1_output_20_partition")

    def test_il_agg_sample_mean_lec_1_partition(self):
        self.genbash("il_agg_sample_mean_lec_1_output", 1)
        self.check("il_agg_sample_mean_lec_1_output_1_partition")

    def test_il_agg_sample_mean_lec_20_partition(self):
        self.genbash("il_agg_sample_mean_lec_1_output", 20)
        self.check("il_agg_sample_mean_lec_1_output_20_partition")

    def test_il_occ_sample_mean_lec_1_partition(self):
        self.genbash("il_occ_sample_mean_lec_1_output", 1)
        self.check("il_occ_sample_mean_lec_1_output_1_partition")

    def test_il_occ_sample_mean_lec_20_partition(self):
        self.genbash("il_occ_sample_mean_lec_1_output", 20)
        self.check("il_occ_sample_mean_lec_1_output_20_partition")

    def test_il_summarycalc_1_partition(self):
        self.genbash("il_summarycalc_1_output", 1)
        self.check("il_summarycalc_1_output_1_partition")

    def test_il_summarycalc_20_partition(self):
        self.genbash("il_summarycalc_1_output", 20)
        self.check("il_summarycalc_1_output_20_partition")

    def test_il_eltcalc_1_partition(self):
        self.genbash("il_eltcalc_1_output", 1)
        self.check("il_eltcalc_1_output_1_partition")

    def test_il_eltcalc_20_partition(self):
        self.genbash("il_eltcalc_1_output", 20)
        self.check("il_eltcalc_1_output_20_partition")

    def test_il_aalcalc_1_partition(self):
        self.genbash("il_aalcalc_1_output", 1)
        self.check("il_aalcalc_1_output_1_partition")

    def test_il_aalcalc_20_partition(self):
        self.genbash("il_aalcalc_1_output", 20)
        self.check("il_aalcalc_1_output_20_partition")

    def test_il_pltcalc_1_partition(self):
        self.genbash("il_pltcalc_1_output", 1)
        self.check("il_pltcalc_1_output_1_partition")

    def test_il_pltcalc_20_partition(self):
        self.genbash("il_pltcalc_1_output", 20)
        self.check("il_pltcalc_1_output_20_partition")

    def test_il_agg_fu_lec_1_partition(self):
        self.genbash("il_agg_fu_lec_1_output", 1)
        self.check("il_agg_fu_lec_1_output_1_partition")

    def test_il_agg_fu_lec_20_partition(self):
        self.genbash("il_agg_fu_lec_1_output", 20)
        self.check("il_agg_fu_lec_1_output_20_partition")

    def test_il_occ_fu_lec_1_output_1_partition(self):
        self.genbash("il_occ_fu_lec_1_output", 1)
        self.check("il_occ_fu_lec_1_output_1_partition")

    def test_il_occ_fu_lec_1_output_20_partition(self):
        self.genbash("il_occ_fu_lec_1_output", 20)
        self.check("il_occ_fu_lec_1_output_20_partition")

    def test_il_agg_ws_lec_1_partition(self):
        self.genbash("il_agg_ws_lec_1_output", 1)
        self.check("il_agg_ws_lec_1_output_1_partition")

    def test_il_agg_ws_lec_20_partition(self):
        self.genbash("il_agg_ws_lec_1_output", 20)
        self.check("il_agg_ws_lec_1_output_20_partition")

    def test_il_occ_ws_lec_1_partition(self):
        self.genbash("il_occ_ws_lec_1_output", 1)
        self.check("il_occ_ws_lec_1_output_1_partition")

    def test_il_occ_ws_lec_20_partition(self):
        self.genbash("il_occ_ws_lec_1_output", 20)
        self.check("il_occ_ws_lec_1_output_20_partition")

    def test_il_agg_ws_mean_lec_1_partition(self):
        self.genbash("il_agg_ws_mean_lec_1_output", 1)
        self.check("il_agg_ws_mean_lec_1_output_1_partition")

    def test_il_agg_ws_mean_lec_20_partition(self):
        self.genbash("il_agg_ws_mean_lec_1_output", 20)
        self.check("il_agg_ws_mean_lec_1_output_20_partition")

    def test_il_occ_ws_mean_lec_1_partition(self):
        self.genbash("il_occ_ws_mean_lec_1_output", 1)
        self.check("il_occ_ws_mean_lec_1_output_1_partition")

    def test_il_occ_ws_mean_lec_20_partition(self):
        self.genbash("il_occ_ws_mean_lec_1_output", 20)
        self.check("il_occ_ws_mean_lec_1_output_20_partition")

    def test_il_agg_sample_mean_lec_1_output_1_partition(self):
        self.genbash("il_agg_sample_mean_lec_1_output", 1)
        self.check("il_agg_sample_mean_lec_1_output_1_partition")

    def test_il_agg_sample_mean_lec_1_output_20_partition(self):
        self.genbash("il_agg_sample_mean_lec_1_output", 20)
        self.check("il_agg_sample_mean_lec_1_output_20_partition")

    def test_il_occ_sample_mean_lec_1_output_1_partition(self):
        self.genbash("il_occ_sample_mean_lec_1_output", 1)
        self.check("il_occ_sample_mean_lec_1_output_1_partition")

    def test_il_occ_sample_mean_lec_1_output_20_partition(self):
        self.genbash("il_occ_sample_mean_lec_1_output", 20)
        self.check("il_occ_sample_mean_lec_1_output_20_partition")

    def test_all_calcs_1_partition(self):
        self.genbash("all_calcs_1_output", 1)
        self.check("all_calcs_1_output_1_partition")

    def test_all_calcs_20_partition(self):
        self.genbash("all_calcs_1_output", 20)
        self.check("all_calcs_1_output_20_partition")

    def test_all_calcs_40_partition(self):
        self.genbash("all_calcs_1_output", 40)
        self.check("all_calcs_1_output_40_partition")

    def test_gul_no_lec_1_output_1_partition(self):
        self.genbash("gul_no_lec_1_output", 1)
        self.check("gul_no_lec_1_output_1_partition")

    def test_gul_no_lec_1_output_2_partition(self):
        self.genbash("gul_no_lec_1_output", 2)
        self.check("gul_no_lec_1_output_2_partition")

    def test_gul_no_lec_2_output_1_partition(self):
        self.genbash("gul_no_lec_2_output", 1)
        self.check("gul_no_lec_2_output_1_partition")

    def test_gul_no_lec_2_output_2_partitions(self):
        self.genbash("gul_no_lec_2_output", 2)
        self.check("gul_no_lec_2_output_2_partition")

    def test_gul_lec_1_output_1_partition(self):
        self.genbash("gul_lec_1_output", 1)
        self.check("gul_lec_1_output_1_partition")

    def test_gul_lec_1_output_2_partitions(self):
        self.genbash("gul_lec_1_output", 2)
        self.check("gul_lec_1_output_2_partition")

    def test_gul_lec_2_output_1_partition(self):
        self.genbash("gul_lec_2_output", 1)
        self.check("gul_lec_2_output_1_partition")

    def test_gul_lec_2_output_2_partitions(self):
        self.genbash("gul_lec_2_output", 2)
        self.check("gul_lec_2_output_2_partition")

    def test_il_no_lec_1_output_1_partition(self):
        self.genbash("il_no_lec_1_output", 1)
        self.check("il_no_lec_1_output_1_partition")

    def test_il_no_lec_1_output_2_partition(self):
        self.genbash("il_no_lec_1_output", 2)
        self.check("il_no_lec_1_output_2_partition")

    def test_il_no_lec_2_output_1_partition(self):
        self.genbash("il_no_lec_2_output", 1)
        self.check("il_no_lec_2_output_1_partition")

    def test_il_no_lec_2_output_2_partitions(self):
        self.genbash("il_no_lec_2_output", 2)
        self.check("il_no_lec_2_output_2_partition")

    def test_il_lec_1_output_1_partition(self):
        self.genbash("il_lec_1_output", 1)
        self.check("il_lec_1_output_1_partition")

    def test_il_lec_1_output_2_partitions(self):
        self.genbash("il_lec_1_output", 2)
        self.check("il_lec_1_output_2_partition")

    def test_il_lec_2_output_1_partition(self):
        self.genbash("il_lec_2_output", 1)
        self.check("il_lec_2_output_1_partition")

    def test_il_lec_2_output_2_partitions(self):
        self.genbash("il_lec_2_output", 2)
        self.check("il_lec_2_output_2_partition")

    def test_gul_il_no_lec_1_output_1_partition(self):
        self.genbash("gul_il_no_lec_1_output", 1)
        self.check("gul_il_no_lec_1_output_1_partition")

    def test_gul_il_no_lec_1_output_2_partition(self):
        self.genbash("gul_il_no_lec_1_output", 2)
        self.check("gul_il_no_lec_1_output_2_partition")

    def test_gul_il_no_lec_2_output_1_partition(self):
        self.genbash("gul_il_no_lec_2_output", 1)
        self.check("gul_il_no_lec_2_output_1_partition")

    def test_gul_il_no_lec_2_output_2_partitions(self):
        self.genbash("gul_il_no_lec_2_output", 2)
        self.check("gul_il_no_lec_2_output_2_partition")

    def test_gul_il_lec_1_output_1_partition(self):
        self.genbash("gul_il_lec_1_output", 1)
        self.check("gul_il_lec_1_output_1_partition")

    def test_gul_il_lec_1_output_2_partitions(self):
        self.genbash("gul_il_lec_1_output", 2)
        self.check("gul_il_lec_1_output_2_partition")

    def test_gul_il_lec_2_output_1_partition(self):
        self.genbash("gul_il_lec_2_output", 1)
        self.check("gul_il_lec_2_output_1_partition")

    def test_gul_il_lec_2_output_2_partitions(self):
        self.genbash("gul_il_lec_2_output", 2)
        self.check("gul_il_lec_2_output_2_partition")

    def test_gul_il_lec_2_output_10_partitions(self):
        self.genbash("gul_il_lec_2_output", 10)
        self.check("gul_il_lec_2_output_10_partition")

    def test_analysis_settings_1(self):
        self.genbash("analysis_settings_1", 1)
        self.check("analysis_settings_1_1_partition")

    def test_analysis_settings_2(self):
        self.genbash("analysis_settings_2", 1)
        self.check("analysis_settings_2_1_partition")

    def test_analysis_settings_3_0_reins_iters(self):
        self.genbash("analysis_settings_3", 1, 1)
        self.check("analysis_settings_3_1_reins_layer_1_partition")

    def test_analysis_settings_4_0_reins_iters(self):
        self.genbash("analysis_settings_4", 1, 1)
        self.check("analysis_settings_4_1_reins_layer_1_partition")



import json
import string
from argparse import Namespace
from tempfile import NamedTemporaryFile
from unittest import TestCase

import os
from hypothesis import given
from hypothesis.strategies import text

from oasislmf.cmd.base import InputValues
from oasislmf.utils.exceptions import OasisException


class InputValuesGet(TestCase):
    @given(text(min_size=1, max_size=10, alphabet=string.ascii_letters))
    def test_variable_is_on_the_command_line_but_not_in_config___command_line_is_used(self, cmd_var):
        with NamedTemporaryFile('w') as conf_file:
            json.dump({'bar': 'boo'}, conf_file)
            conf_file.flush()

            args = Namespace(foo=cmd_var, config=conf_file.name)

            inputs = InputValues(args)

            self.assertEqual(cmd_var, inputs.get('foo'))

    @given(
        text(min_size=1, max_size=10, alphabet=string.ascii_letters),
        text(min_size=1, max_size=10, alphabet=string.ascii_letters),
    )
    def test_variable_is_on_the_command_line_and_in_config___command_line_is_used(self, cmd_var, conf_var):
        with NamedTemporaryFile('w') as conf_file:
            json.dump({'foo': conf_var}, conf_file)
            conf_file.flush()

            args = Namespace(foo=cmd_var, config=conf_file.name)

            inputs = InputValues(args)

            self.assertEqual(cmd_var, inputs.get('foo'))

    @given(text(min_size=1, max_size=10, alphabet=string.ascii_letters))
    def test_variable_is_not_on_the_command_line_but_is_in_config___config_is_used(self, conf_var):
        with NamedTemporaryFile('w') as conf_file:
            json.dump({'foo': conf_var}, conf_file)
            conf_file.flush()

            args = Namespace(config=conf_file.name)

            inputs = InputValues(args)

            self.assertEqual(conf_var, inputs.get('foo'))

    def test_variable_is_not_on_the_command_line_or_config_var_is_not_required___none_is_returned(self):
        with NamedTemporaryFile('w') as conf_file:
            json.dump({'bar': 'boo'}, conf_file)
            conf_file.flush()

            args = Namespace(config=conf_file.name)

            inputs = InputValues(args)

            self.assertIsNone(inputs.get('foo'))

    @given(text(min_size=1, max_size=10, alphabet=string.ascii_letters))
    def test_variable_is_not_on_the_command_line_or_config_var_is_not_required_default_is_supplied___default_is_returned(self, default):
        with NamedTemporaryFile('w') as conf_file:
            json.dump({'bar': 'boo'}, conf_file)
            conf_file.flush()

            args = Namespace(config=conf_file.name)

            inputs = InputValues(args)

            self.assertEqual(default, inputs.get('foo', default=default))

    def test_variable_is_not_on_the_command_line_or_config_var_is_required___error_is_raised(self):
        with NamedTemporaryFile('w') as conf_file:
            json.dump({'bar': 'boo'}, conf_file)
            conf_file.flush()

            args = Namespace(config=conf_file.name)

            inputs = InputValues(args)

            with self.assertRaises(OasisException):
                inputs.get('foo', required=True)

    def test_variable_is_a_path___path_is_relative_to_config_file(self):
        with NamedTemporaryFile('w') as conf_file:
            json.dump({'foo': './some/path'}, conf_file)
            conf_file.flush()

            expected_result = os.path.abspath(
                os.path.join(os.path.dirname(conf_file.name), 'some', 'path')
            )

            args = Namespace(config=conf_file.name)

            inputs = InputValues(args)

            result = inputs.get('foo', required=True, is_path=True)

            self.assertEqual(expected_result, result)



# -*- coding: utf-8 -*-

__all__ = [
    'Translator'
]

import json
import logging
import multiprocessing
import os

import pandas as pd

from lxml import etree

from oasislmf.utils.concurrency import (
    multiprocess,
    multithread,
    Task,
)

class Translator(object):
    def __init__(self, input_path, output_path, xslt_path, xsd_path=None, append_row_nums=False, chunk_size=5000, logger=None):
        """
        Transforms exposures/locations in CSV format
        by converting a source file to XML and applying an XSLT transform
        to apply rules for selecting, merging or updating columns

        An optional step is to passing an XSD file for output validation

        :param input_path: Source exposures file path, which should be in CSV comma delimited format
        :type input_path: str

        :param output_path: File to write transform results
        :type output_path: str

        :param xslt_path: Source exposures Transformation rules file
        :type xslt_path: str

        :param xsd_path: Source exposures output validation file
        :type xsd_path: str

        :param append_row_nums: Append line numbers to first column of output called `ROW_ID` [1 .. n] when n is the number of rows processed.
        :type append_row_nums: boolean

        :param chunk_size: Number of rows to process per multiprocess Task
        :type chunk_size: int
        """

        self.logger = logger or logging.getLogger()
        self.xsd = (etree.parse(xsd_path) if xsd_path else None)
        self.xslt = etree.parse(xslt_path)
        self.fpath_input = input_path
        self.fpath_output = output_path

        self.row_nums = append_row_nums
        self.row_limit = chunk_size
        self.row_header_in = None
        self.row_header_out = None

    def __call__(self):
        csv_reader = pd.read_csv(self.fpath_input, iterator=True, dtype=object, encoding='utf-8')

        task_list = []
        for chunk_id, (data, first_row, last_row) in enumerate(self.next_file_slice(csv_reader)):
            task_list.append(Task(self.process_chunk, args=(data,first_row,last_row, chunk_id), key=chunk_id))

        results = {}
        num_ps = multiprocessing.cpu_count()
        for key, data in multithread(task_list, pool_size=num_ps):
            results[key] = data

        ## write output to disk
        for i in range(0, len(results)):
            if (i == 0):
                self.write_file_header(results[i].columns.tolist())
            results[i].to_csv(
                self.fpath_output,
                mode='a',
                encoding='utf-8',
                header=False,
                index=False,
            )

    def process_chunk(self, data, first_row_number, last_row_number, seq_id):
        xml_input_slice = self.csv_to_xml(
            self.row_header_in,
            data
        )
        self.print_xml(xml_input_slice)

        # Transform
        xml_output = self.xml_transform(xml_input_slice, self.xslt)

        # Validate Output
        if self.xsd:
            self.logger.debug(self.xml_validate(xml_output, self.xsd))

        # Convert transform XML back to CSV
        return self.xml_to_csv(
                 xml_output,        # XML etree
                 first_row_number,  # First Row in this slice
                 last_row_number    # Last Row in this slice
        )


    def csv_to_xml(self, csv_header, csv_data):
        root = etree.Element('root')
        for row in csv_data:
            rec = etree.SubElement(root, 'rec')
            for i in range(0, len(row)):
                if(row[i] not in [None, "", 'NaN']):
                    rec.set(csv_header[i], row[i])
        return root

    def xml_to_csv(self, xml_elementTree, row_first, row_last):
        root = xml_elementTree.getroot()
        # Set output col headers
        if not (self.row_header_out):
            self.row_header_out = root[0].keys()

        # create Dataframe from xml and append each row
        rows = []
        for rec in root:
            rows.append(dict(rec.attrib))

        df_out = pd.DataFrame(rows, columns=self.row_header_out)

        # Add column for row_nums if set
        if self.row_nums:
            start = row_first + 1
            end = start + len(df_out)
            df_out.insert(0, 'ROW_ID', pd.Series(range(start,end)))

        return df_out

    def xml_validate(self, xml_etree, xsd_etree):
        xmlSchema = etree.XMLSchema(xsd_etree)
        self.print_xml(xml_etree)
        self.print_xml(xsd_etree)
        if (xmlSchema.validate(xml_etree)):
            return True
        else:
            if self.logger.isEnabledFor(logging.DEBUG):
                self.logger.error('Input failed to Validate')
                log = xmlSchema.error_log
                self.logger.error(log.last_error)
            return False

    def xml_transform(self, xml_doc, xslt):
        lxml_transform = etree.XSLT(xslt)
        return lxml_transform(xml_doc)

    def next_file_slice(self, file_reader):
        while True:
            try:
                df_slice = file_reader.get_chunk(self.row_limit)
                if(not self.row_header_in):
                    self.row_header_in = df_slice.columns.values.tolist()
                yield (
                    df_slice.fillna("").values.astype("unicode").tolist(),
                    df_slice.first_valid_index(),
                    df_slice.last_valid_index()
                )
            except StopIteration:
                self.logger.debug('End of input file')
                break

    def write_file_header(self, row_names):
        pd.DataFrame(columns=row_names).to_csv(
            self.fpath_output,
            encoding='utf-8',
            header=True,
            index=False,
        )

    def print_xml(self, etree_obj):
        self.logger.debug('___________________________________________')
        self.logger.debug(etree.tostring(etree_obj, pretty_print=True))



from __future__ import unicode_literals

import six

INPUT_FILES = {
    'items': {'name': 'items', 'type': 'gul', 'conversion_tool': 'itemtobin'},
    'coverages': {'name': 'coverages', 'type': 'gul', 'conversion_tool': 'coveragetobin'},
    'gulsummaryxref': {'name': 'gulsummaryxref', 'type': 'gul', 'conversion_tool': 'gulsummaryxreftobin'},
    'events': {'name': 'events', 'type': 'optional', 'conversion_tool': 'evetobin'},
    'fm_policytc': {'name': 'fm_policytc', 'type': 'il', 'conversion_tool': 'fmpolicytctobin'},
    'fm_profile': {'name': 'fm_profile', 'type': 'il', 'conversion_tool': 'fmprofiletobin'},
    'fm_programme': {'name': 'fm_programme', 'type': 'il', 'conversion_tool': 'fmprogrammetobin'},
    'fm_xref': {'name': 'fm_xref', 'type': 'il', 'conversion_tool': 'fmxreftobin'},
    'fmsummaryxref': {'name': 'fmsummaryxref', 'type': 'il', 'conversion_tool': 'fmsummaryxreftobin'}
}
GUL_INPUT_FILES = {k: v for k, v in six.iteritems(INPUT_FILES) if v['type'] == 'gul'}
IL_INPUT_FILES = {k: v for k, v in six.iteritems(INPUT_FILES) if v['type'] == 'il'}
OPTIONAL_INPUT_FILES = {k: v for k, v in six.iteritems(INPUT_FILES) if v['type'] == 'optional'}

TAR_FILE = 'inputs.tar.gz'

GENERAL_SETTINGS_FILE = "general_settings.csv"
MODEL_SETTINGS_FILE = "model_settings.csv"
GUL_SUMMARIES_FILE = "gul_summaries.csv"
IL_SUMMARIES_FILE = "il_summaries.csv"



from __future__ import unicode_literals

import string
from unittest import TestCase

import os
import io

import six
from backports.tempfile import TemporaryDirectory
from hypothesis import given
from hypothesis.strategies import text

from oasislmf.cmd import RootCmd
from oasislmf import __version__


class GenerateModelTesterDockerfileRun(TestCase):
    def get_command(self, api_server_url='http://localhost:8001', extras=None):
        kwargs_str = ' '.join('--{} {}'.format(k, v) for k, v in six.iteritems(extras or {}))

        return RootCmd(argv='test gen-model-tester-dockerfile {} {}'.format(kwargs_str, api_server_url).split())

    def test_api_server_is_not_supplied___parsing_args_raises_an_error(self):
        cmd = self.get_command(api_server_url='')

        with self.assertRaises(SystemExit):
            cmd.parse_args()

    def test_model_data_directory_is_not_supplied___default_is_abs_path_to_model_data_directory(self):
        cmd = self.get_command()

        args = cmd.parse_args()

        self.assertEqual(args.model_data_directory, os.path.abspath('./model_data'))

    def test_model_data_directory_is_supplied___model_data_directory_abs_path_to_supplied_path(self):
        cmd = self.get_command(extras={'model-data-directory': './other_model_data'})

        args = cmd.parse_args()

        self.assertEqual(args.model_data_directory, os.path.abspath('./other_model_data'))

    def test_model_version_is_not_supplied___default_value_is_none(self):
        cmd = self.get_command()

        args = cmd.parse_args()

        self.assertIsNone(args.model_version_file)

    def test_model_version_paths_supplied___version_path_is_abs_path_to_version_file(self):
        cmd = self.get_command(extras={'model-version-file': './specific_version_file.csv'})

        args = cmd.parse_args()

        self.assertEqual(args.model_version_file, os.path.abspath('./specific_version_file.csv'))

    @given(
        text(alphabet=string.ascii_letters, min_size=1),
        text(alphabet=string.ascii_letters, min_size=1),
        text(alphabet=string.ascii_letters, min_size=1),
        text(alphabet=string.ascii_letters, min_size=1),
    )
    def test_version_file_is_not_specified___version_file_from_data_directory_is_used(self, server_url, org, model, model_version):
        with TemporaryDirectory() as d:
            with io.open(os.path.join(d, 'ModelVersion.csv'), 'w', encoding='utf-8') as version_file:
                version_file.write('{},{},{}'.format(org, model, model_version))

            with io.open(os.path.join(os.path.dirname(__file__), os.path.pardir, os.path.pardir, 'oasislmf', '_data', 'Dockerfile.model_api_tester'), encoding='utf-8') as tpl_file:
                tpl_content = tpl_file.read()
                expected = tpl_content.replace(
                    '%CLI_VERSION%', __version__,
                ).replace(
                    '%OASIS_API_SERVER_URL%', server_url,
                ).replace(
                    '%SUPPLIER_ID%', org,
                ).replace(
                    '%MODEL_ID%', model,
                ).replace(
                    '%MODEL_VERSION%', model_version,
                ).replace(
                    '%LOCAL_MODEL_DATA_PATH%', d,
                )

            cmd = self.get_command(api_server_url=server_url, extras={'model-data-directory': d})
            res = cmd.run()

            with io.open(os.path.join(d, 'Dockerfile.{}_{}_model_api_tester'.format(org.lower(), model.lower())), encoding='utf-8') as docker_file:
                docker_content = docker_file.read()

            self.assertEqual(0, res)
            self.assertEqual(expected, docker_content)

    @given(
        text(alphabet=string.ascii_letters, min_size=1),
        text(alphabet=string.ascii_letters, min_size=1),
        text(alphabet=string.ascii_letters, min_size=1),
        text(alphabet=string.ascii_letters, min_size=1),
    )
    def test_version_file_is_specified___specified_version_is_used(self, server_url, org, model, model_version):
        with TemporaryDirectory() as d:
            with io.open(os.path.join(d, 'other_version_file.csv'), 'w', encoding='utf-8') as version_file:
                version_file.write('{},{},{}'.format(org, model, model_version))

            with io.open(os.path.join(os.path.dirname(__file__), os.path.pardir, os.path.pardir, 'oasislmf', '_data', 'Dockerfile.model_api_tester'), encoding='utf-8') as tpl_file:
                tpl_content = tpl_file.read()
                expected = tpl_content.replace(
                    '%CLI_VERSION%', __version__,
                ).replace(
                    '%OASIS_API_SERVER_URL%', server_url,
                ).replace(
                    '%SUPPLIER_ID%', org,
                ).replace(
                    '%MODEL_ID%', model,
                ).replace(
                    '%MODEL_VERSION%', model_version,
                ).replace(
                    '%LOCAL_MODEL_DATA_PATH%', d,
                )

            cmd = self.get_command(api_server_url=server_url, extras={
                'model-data-directory': d,
                'model-version-file': os.path.join(d, 'other_version_file.csv')
            })
            res = cmd.run()

            with io.open(os.path.join(d, 'Dockerfile.{}_{}_model_api_tester'.format(org.lower(), model.lower())), encoding='utf-8') as docker_file:
                docker_content = docker_file.read()

            self.assertEqual(0, res)
            self.assertEqual(expected, docker_content)



# -*- coding: utf-8 -*-

"""
    Utilities for running system or file-related commands, and other OS-related utilities.
"""
import io
import socket

from .exceptions import OasisException

__all__ = [
    'load_ini_file',
    'replace_in_file',
]


def load_ini_file(ini_file_path):
    """
    Reads an INI file and returns it as a dictionary.
    """
    try:
        with io.open(ini_file_path, 'r', encoding='utf-8') as f:
            lines = [
                l.strip() for l in [l for l in f.read().split('\n') if l and not l.startswith('[')]
            ]
    except IOError as e:
        raise OasisException(str(e))

    di = dict(map(lambda kv: (kv[0].strip(), kv[1].strip()), (line.split('=') for line in lines)))

    for k in di:
        if di[k].lower() == 'true':
            di[k] = True
        elif di[k].lower() == 'false':
            di[k] = False
        else:
            ipf = lambda s: socket.inet_ntoa(socket.inet_aton(s))
            for conv in (int, float, ipf, str):
                try:
                    di[k] = conv(di[k])
                    break
                except:  # noqa: 722
                    continue
    return di


def replace_in_file(source_file_path, target_file_path, var_names, var_values):
    """
    Replaces a list of placeholders / variable names in a source file with a
    matching set of values, and writes it out to a new target file.
    """
    if len(var_names) != len(var_values):
        raise OasisException('Number of variable names does not equal the number of variable values to replace - please check and try again.')

    try:
        with io.open(source_file_path, 'r') as f:
            lines = f.readlines()

        with io.open(target_file_path, 'w') as f:
            for i in range(len(lines)):
                outline = inline = lines[i]

                present_var_names = [v for v in var_names if v in inline]

                if present_var_names:
                    for var_name in present_var_names:
                        var_value = var_values[var_names.index(var_name)]
                        outline = outline.replace(var_name, var_value)
                f.write(outline)
    except (OSError, IOError) as e:
        raise OasisException(str(e))
